#include "sparseMatrix.hpp"

// ================================================================
// Public members
// ================================================================

// Constructors

SparseMatrix::SparseMatrix(const int M, const int N) {

  _rows.resize(N);
  _cols.resize(N);
  _vals.resize(N);

  _size = M;
  _nNz  = N;

}

SparseMatrix::SparseMatrix(const SparseMatrix& that) {

  _rows = that._rows;
  _cols = that._cols;
  _vals = that._vals;

  _size = that._size;
  _nNz  = that._nNz;

}

SparseMatrix::SparseMatrix(const std::string fileName) {

  std::ifstream mFile(fileName.c_str());

  if (not mFile.good()) {
    std::cerr << "Error while opening " << fileName << std::endl;
    abort();
  }
  
  _size = 0;

  mFile >> _nNz;
  _rows.resize(_nNz);
  _cols.resize(_nNz);
  _vals.resize(_nNz);
  
  for (uint i=0u; i<_nNz; i++) {
    mFile >> _rows[i] >> _cols[i] >> _vals[i];
    if (_rows[i] > _size) _size = _rows[i];
    if (_cols[i] > _size) _size = _cols[i];
  }

  _size++; // Indices start at 0!

}

// Other functions

double SparseMatrix::getValue(const int i, const int j) {

  int pos = posIndex(i,j);
  return pos==-1 ? 0. : _vals[pos];

}

double SparseMatrix::inftNorm() {

  if (_nNz==0) {
    std::cout << "Warning: empty matrix !" << std::endl;
    return 0.;
  }  
 
  // Compute sums
  std::vector<double> maxs(_size,0.);
  for (uint i=0u; i<_nNz; i++)
    maxs[_rows[i]] += abs(_vals[i]);

  // Get max of sums
  double res = maxs[0];
  for (uint i=1u; i<_size; i++)
    if (maxs[i]>res) res = maxs[i];
  return res;

}

double SparseMatrix::trace() {
  double res = 0.e0;
  for (uint i=0u; i<_nNz; i++)
    if (_rows[i]==_cols[i])
      res += _vals[i];
  return res;
}

void SparseMatrix::mult(const rvec& x, rvec& res) {

  // Check matrix and vector shapes
  if (x.size()>_size)
    std::cout << "matMul: Warning, will only use " << _size << " first coefficients of x (size of matrix)." << std::endl;
  else if (x.size()<_size) {
    std::cerr << "matMul: Vector (" << x.size() << ") is smaller than matrix (" << _size << ")." << std::endl;
    abort();
  }

  // (re)initialize result
  res.clear();
  for (uint i=0u; i<x.size(); i++)
    res.push_back(0.);

  // Add each matrix coefficient contribution
  for (uint i=0u; i<_nNz; i++)
    res[_rows[i]] += _vals[i]*x[_cols[i]];
}

double& SparseMatrix::operator()(const int i, const int j) {

  if (i>= _size or j>= _size) {
    std::cerr << "Index (" << i << "," << j << ") is out of matrix bounds, size: " << _size << std::endl;
    abort();
  }

  int pos = posIndex(i,j);

  if (pos == -1) {
    _rows.push_back(i);
    _cols.push_back(j);
    _vals.push_back(0.);
    pos = _vals.size()-1;
    _nNz++;
  } 

  return _vals[pos];

}

// Display
void SparseMatrix::print(std::ostream& os) {

  os << "Sparse matrix: size: " << _size << ", non-zero terms (" << _nNz << "):" << std::endl;
  for (uint i=0u; i<_nNz; i++)
    os << "(" << _rows[i] << "," << _cols[i] << ") : " << _vals[i] << std::endl;

}

// ================================================================
// Private members
// ================================================================

int SparseMatrix::posIndex(const int i, const int j) {

  int res = -1;
  int k   = 0;
  do {
    if (_rows[k]==i and _cols[k]==j)
      res = k; 
    k++;
  } while (res==-1 and k<_nNz);
  return res;

}


// ================================================================
// Operators
// ================================================================

std::ostream& operator<<(std::ostream& os, SparseMatrix& M) {
  M.print(os);
  return os;
}

SparseMatrix operator+(const SparseMatrix& A, const SparseMatrix& B) {
  
  SparseMatrix res(A); // Start by copying A

  for (int i=0; i<B._nNz; i++)
    res(B._rows[i], B._cols[i]) += B._vals[i];

  return res;

}


